@@ -6,13 +6,13 @@
 
 @license: MIT
 
-@version: 0.1.5
+@version: 0.1.7
 
 @email: owuordickson@gmail.com
 
 @created: 21 July 2021
 
-@modified: 01 March 2022
+@modified: 17 March 2022
 
 SO4GP
 ------
@@ -20,10 +20,11 @@ SO4GP
 **SO4GP** stands for: "Some Optimizations for Gradual Patterns". SO4GP applies optimizations such as swarm intelligence,
 HDF5 chunks, SVD and many others in order to improve the efficiency of extracting gradual patterns (GPs).
 
- A GP is a set of gradual items (GI) and its quality is measured by its computed support value. For example
- given a data set with 3 columns (age, salary, cars) and 10 objects. A GP may take the form: {age+, salary-} with a
- support of 0.8. This implies that 8 out of 10 objects have the values of column age 'increasing' and column 'salary'
- decreasing.
+ A GP is a set of gradual items (GI) and its quality is measured by its computed support value. A GI is a pair (i,v)
+ where i is a column and v is a variation symbol: increasing/decreasing. Each column of a data set yields 2 GIs; for
+ example, column age yields GI age+ or age-. For example given a data set with 3 columns (age, salary, cars) and 10
+ objects. A GP may take the form: {age+, salary-} with a support of 0.8. This implies that 8 out of 10 objects have the
+ values of column age 'increasing' and column 'salary' decreasing.
 
     The classical approach for mining GPs is computationally expensive. This package provides Python algorithm
 implementations of several optimization techniques that are applied to the classical approach in order to improve its
@@ -45,11 +46,15 @@ import time
 import gc
 import numpy as np
 import json
+import matplotlib.pyplot as plt
+import multiprocessing as mp
+import os
 import pandas as pd
 import random
 from ypstruct import structure
 
-# -------- CONFIGURATION (START)-------------
+
+# -------- CONFIGURATION -------------
 
 # Global Swarm Configurations
 MIN_SUPPORT = 0.5
@@ -78,10 +83,7 @@ N_PARTICLES = 5
 STEP_SIZE = 0.5
 
 
-# -------- CONFIGURATION (END)-------------
-
-
-# -------- DATA SET PREPARATION (START)-------------
+# -------- DATA SET PREPARATION -------------
 
 """
 Changes
@@ -156,6 +158,8 @@ class DataGP:
 
             valid_bins: valid bitmaps (in the form of ndarray) of all gradual items corresponding to the attr_cols, a bitmap is valid if its computed support is equal or greater than the minimum support threshold
 
+            net_wins: a net-wins matrix constructed from valid gradual item bitmaps
+
             no_bins: True if all none of the attr_cols yields a valid bitmap
 
         :param data_source: [required] a numeric data source, it can either be a 'file in csv format' or a 'Pandas DataFrame'
@@ -172,8 +176,6 @@ class DataGP:
         """:type titles: ndarray"""
         """:type data: ndarray"""
         self.row_count, self.col_count = self.data.shape
-        """:type row_count: int"""
-        """:type col_count: int"""
         self.time_cols = self.get_time_cols()
         self.attr_cols = self.get_attr_cols()
         self.valid_bins = np.array([])
@@ -212,8 +214,8 @@ class DataGP:
 
     def init_attributes(self, attr_data=None):
         """
-        Generates bitmaps for columns with numeric objects. It only stores (attribute valid_bins) those bitmaps whose computed support values
-        are greater or equal to the minimum support threshold value.
+        Generates bitmaps for columns with numeric objects. It only stores (attribute valid_bins) those bitmaps whose
+        computed support values are greater or equal to the minimum support threshold value.
 
         :param attr_data: stepped attribute objects
         :return: void
@@ -386,10 +388,115 @@ def gibitmap(d_set, col):
             temp_pos = np.where(col_data < col_data[:, np.newaxis], 1, 0)
         return temp_pos
 
-# -------- DATA SET PREPARATION (END)-------------
+
+# -------- PROFILE PERFORMANCE -----------
+
+class Profile:
+
+    @staticmethod
+    def get_num_cores():
+        num_cores = Profile.get_slurm_cores()
+        if not num_cores:
+            num_cores = mp.cpu_count()
+        return num_cores
+
+    @staticmethod
+    def get_slurm_cores():
+        try:
+            cores = int(os.environ['SLURM_JOB_CPUS_PER_NODE'])
+            return cores
+        except ValueError:
+            try:
+                str_cores = str(os.environ['SLURM_JOB_CPUS_PER_NODE'])
+                temp = str_cores.split('(', 1)
+                cpus = int(temp[0])
+                str_nodes = temp[1]
+                temp = str_nodes.split('x', 1)
+                str_temp = str(temp[1]).split(')', 1)
+                nodes = int(str_temp[0])
+                cores = cpus * nodes
+                return cores
+            except ValueError:
+                return False
+        except KeyError:
+            return False
+
+    @staticmethod
+    def get_quick_mem_block(snapshot):
+        top_stats = snapshot.statistics('traceback')
+
+        # pick the biggest memory block
+        stat = top_stats[0]
+        wr_line = ("%s memory blocks: %.1f KiB" % (stat.count, stat.size / 1024))
+        return wr_line
+
+    @staticmethod
+    def get_quick_mem_use(snapshot, key_type='lineno'):
+        import tracemalloc
+        snapshot = snapshot.filter_traces((
+            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
+            tracemalloc.Filter(False, "<unknown>"),
+        ))
+        top_stats = snapshot.statistics(key_type)
+        total = sum(stat.size for stat in top_stats)
+        wr_line = ("Total allocated memory size: %.1f KiB" % (total / 1024))
+        return wr_line
+
+    @staticmethod
+    def get_detailed_mem_use(snapshot, key_type='lineno', limit=30):
+        import tracemalloc
+        import linecache
+        wr_line = ""
+        snapshot = snapshot.filter_traces((
+            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
+            tracemalloc.Filter(False, "<unknown>"),
+        ))
+        top_stats = snapshot.statistics(key_type)
+
+        wr_line += ("Top %s lines" % limit)
+        for index, stat in enumerate(top_stats[:limit], 1):
+            frame = stat.traceback[0]
+            # replace "/path/to/module/file.src" with "module/file.src"
+            filename = os.sep.join(frame.filename.split(os.sep)[-2:])
+            wr_line += ("\n #%s: %s:%s: %.1f KiB" % (index, filename, frame.lineno, stat.size / 1024))
+            line = linecache.getline(frame.filename, frame.lineno).strip()
+            if line:
+                wr_line += ('\n    %s' % line)
+
+        other = top_stats[limit:]
+        if other:
+            size = sum(stat.size for stat in other)
+            wr_line += ("\n %s other: %.1f KiB" % (len(other), size / 1024))
+        total = sum(stat.size for stat in top_stats)
+        wr_line += ("\n Total allocated size: %.1f KiB" % (total / 1024))
+        return wr_line
+
+    @staticmethod
+    def write_file(data, path, wr=True):
+        if wr:
+            with open(path, 'w') as f:
+                f.write(data)
+                f.close()
+        else:
+            pass
+
+    @staticmethod
+    def plot_curve(out, title, pl=False):
+        if pl:
+            # Results
+            plt.plot(out.best_costs)
+            # DO NOT ADD *** plt.semilogy(out.best_costs)
+            plt.xlim(0, out.max_iteration)
+            plt.xlabel('Iterations')
+            plt.ylabel('Best Cost')
+            plt.title(title)
+            plt.grid(True)
+            plt.show()
+        else:
+            pass
 
 
-# -------- GRADUAL PATTERNS (START)-------------
+# -------- GRADUAL PATTERNS -------------
 
 """
 @created: "20 May 2020"
@@ -404,7 +511,7 @@ GP: Gradual Pattern {(0, +), (1, -), (3, +)}
 class GI:
     """Description of class GI (Gradual Item)
 
-    A class that is used to create GI objects. A GI is a pair (i,v) where i a column and v is a variation symbola -
+    A class that is used to create GI objects. A GI is a pair (i,v) where i is a column and v is a variation symbol -
     increasing/decreasing. Each column of a data set yields 2 GIs; for example, column age yields GI age+ or age-.
 
     This class has the following attributes:
@@ -412,7 +519,7 @@ class GI:
 
         symbol: variation symbol (either + or -)
 
-        gradual_item: the GI in an ndarray format
+        gradual_item: the GI in a ndarray format
 
         tuple: the GI in a tuple format
 
@@ -423,7 +530,7 @@ class GI:
     def __init__(self, attr_col, symbol):
         """Description of class GI (Gradual Item)
 
-        A class that is used to create GI objects. A GI is a pair (i,v) where i a column and v is a variation symbola -
+        A class that is used to create GI objects. A GI is a pair (i,v) where is a column and v is a variation symbol -
         increasing/decreasing. Each column of a data set yields 2 GIs; for example, column age yields GI age+ or age-.
 
         This class has the following attributes:
@@ -431,7 +538,7 @@ class GI:
 
             symbol: variation symbol (either + or -)
 
-            gradual_item: the GI in an ndarray format
+            gradual_item: the GI in a ndarray format
 
             tuple: the GI in a tuple format
 
@@ -643,7 +750,7 @@ class GP:
         """
         Checks if a gradual item (GI) is a member of a gradual pattern (GP)
         :param gi: gradual item
-        :return: True if its a member, otherwise False
+        :return: True if it is a member, otherwise False
         """
         if gi is None:
             return False
@@ -651,6 +758,19 @@ class GP:
             return True
         return False
 
+    def contains_strict(self, gi):
+        """
+        Strictly checks if a gradual item (GI) is a member of a gradual pattern (GP)
+        :param gi: gradual item
+        :return: True if it is a member, otherwise False
+        """
+        if gi is None:
+            return False
+        for gi_obj in self.gradual_items:
+            if (gi.attribute_col == gi_obj.attribute_col) and (gi.symbol == gi_obj.symbol):
+                return True
+        return False
+
     def contains_attr(self, gi):
         """
         Checks is any gradual item (GI) in the gradual pattern (GP) is composed of the column
@@ -759,13 +879,8 @@ class TimeLag:
         return txt
 
 
-# -------- GRADUAL PATTERNS (START)-------------
+# --------- GRAANK ---------------------
 
-
-# --------- GRAANK (START) ---------------------
-
-
-# --------- GRAANK (END) --------------------
 """
 CHANGES:
 1. Removed T-GRAANK modification
@@ -896,7 +1011,7 @@ def graank(f_path=None, min_sup=MIN_SUPPORT, eq=False, return_gps=False):
         return out
 
 
-# -------- ACO-GRAD (START)-------------
+# -------- ACO-GRAD -------------
 
 """
 CHANGES:
@@ -950,7 +1065,7 @@ def acogps(f_path, min_supp=MIN_SUPPORT, evaporation_factor=EVAPORATION_FACTOR,
      candidates are generated.
 
     :param f_path: [required] a numeric data source, it can either be a 'file in csv format' or a 'Pandas DataFrame'
-    :param min_sup: [optional] minimum support threshold, the default is 0.5
+    :param min_supp: [optional] minimum support threshold, the default is 0.5
     :param evaporation_factor: [optional] evaporation factor default = 0.5
     :param max_iteration: [optional] maximum iterations default = 1
     :param return_gps: [optional] additionally return object GPs, the default is False. If set to True, the method returns 2 items: JSON object, GP list
@@ -1162,11 +1277,7 @@ def isduplicate(pattern, lst_winners, lst_losers=None):
     return False
 
 
-# -------- ACO-GRAD (END)-------------
-
-
-# -------- GA-GRAD (START)----------
-
+# -------- GA-GRAD ----------
 
 """
 @author: "Dickson Owuor"
@@ -1201,7 +1312,7 @@ def gagps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, n_pop=N_
      candidates) and find those with the lowest cost as efficiently as possible.
 
     :param data_src: [required] a numeric data source, it can either be a 'file in csv format' or a 'Pandas DataFrame'
-    :param min_sup: [optional] minimum support threshold, the default is set to 0.5
+    :param min_supp: [optional] minimum support threshold, the default is set to 0.5
     :param max_iteration: [optional] maximum iterations, the default is set to 1
     :param n_pop: [optional] initial population, the default is set to 5
     :param pc: [optional] offspring population multiple, the default is set 0.5. This determines how fast the population grows.
@@ -1472,11 +1583,7 @@ def decodegp(attr_keys, position):
     return temp_gp
 
 
-# -------- GA-GRAD (END)-------------
-
-
-# -------- PSO-GRAD (START)----------
-
+# -------- PSO-GRAD ----------
 
 """
 @author: "Dickson Owuor"
@@ -1510,7 +1617,7 @@ def psogps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, n_parti
      those with the highest fitness as efficiently as possible.
 
     :param data_src: [required] a numeric data source, it can either be a 'file in csv format' or a 'Pandas DataFrame'
-    :param min_sup: [optional] minimum support threshold, the default is set to 0.5
+    :param min_supp: [optional] minimum support threshold, the default is set to 0.5
     :param max_iteration: [optional] maximum iterations, the default is set to 1
     :param n_particles: [optional] initial particle population, the default is set to 5
     :param velocity: [optional] particle velocity, the default is set to 0.9
@@ -1626,11 +1733,7 @@ def psogps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, n_parti
         return out
 
 
-# -------- PSO-GRAD (END)-------------
-
-
-# -------- PLS-GRAD (START)----------
-
+# -------- PLS-GRAD ----------
 
 """
 @author: "Dickson Owuor"
@@ -1663,7 +1766,7 @@ def hcgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, step_siz
      cost as efficiently as possible.
 
     :param data_src: [required] a numeric data source, it can either be a 'file in csv format' or a 'Pandas DataFrame'
-    :param min_sup: [optional] minimum support threshold, the default is set to 0.5
+    :param min_supp: [optional] minimum support threshold, the default is set to 0.5
     :param max_iteration: [optional] maximum iterations, the default is set to 1
     :param step_size: [optional] step size, the default is set to 0.5
     :param return_gps: [optional] additionally return object GPs, the default is False. If set to True, the method returns 2 items: JSON object, GP list
@@ -1753,11 +1856,7 @@ def hcgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, step_siz
         return out
 
 
-# -------- PLS-GRAD (END)-------------
-
-
-# -------- PRS-GRAD (START)----------
-
+# -------- PRS-GRAD ----------
 
 """
 @author: "Dickson Owuor"
@@ -1789,7 +1888,7 @@ def rsgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, return_g
      cost as efficiently as possible.
 
     :param data_src: [required] a numeric data source, it can either be a 'file in csv format' or a 'Pandas DataFrame'
-    :param min_sup: [optional] minimum support threshold, the default is set to 0.5
+    :param min_supp: [optional] minimum support threshold, the default is set to 0.5
     :param max_iteration: [optional] maximum iterations, the default is set to 1
     :param return_gps: [optional] additionally return object GPs, the default is False. If set to True, the method returns 2 items: JSON object, GP list
     :return: JSON object
@@ -1872,5 +1971,3 @@ def rsgps(data_src, min_supp=MIN_SUPPORT, max_iteration=MAX_ITERATIONS, return_g
         return out, best_patterns
     else:
         return out
-
-# -------- PRS-GRAD (END)-------------
